--[[
--
-- Makes the game's adjectives appear as suffixes.
--
-- By @simplex.
--
--]]


local _G = GLOBAL
local assert, error = _G.assert, _G.error
local type = _G.type
local tostring = _G.tostring
local table = _G.table
local getmetatable, setmetatable = _G.getmetatable, _G.setmetatable
local rawget, rawset = _G.rawget, _G.rawset
local Class = _G.Class
local EntityScript = _G.EntityScript


local function IsWordable(x)
	if type(x) == "string" then return true end

	local m = getmetatable(x)
	return m and m.__tostring
end

--[[
--
-- Adjective and Adverb classes.
--
--]]

--[[
-- Together with the pieces put in suffices, it forms a whole sentence (clause, to be precise).
-- Becomes a suffix when the final sentence is formed.
--]]
local Modifier = Class(function(self, str)
	if not IsWordable(str) then
		return error("Attempt to initialize modifier with non-string value `"..tostring(str).."'", 2)
	end
	self.str = tostring(str)

	self.space = nil
	self.suffixes = {}
end)

local function IsModifier(x)
	if type(x) ~= "table" then return false end
	local is_a = rawget(x, "is_a")
	return type(is_a) == "function" and is_a(x, Modifier)
end

function Modifier:Copy()
	local p = getmetatable(self)(self.str)
	p.space = self.space
	for i, v in ipairs(self.suffixes) do
		p.suffixes[i] = v
	end
	return p
end

function Modifier.__concat(a, b)
	if IsModifier(b) then
		return getmetatable(a)(tostring(b)..tostring(a))
	end

	if not IsWordable(b) then
		return error("Attempt to concatenate non-string value `"..tostring(b).."' with modifier `"..tostring(a.str).."'", 2)
	end

	b = tostring(b)

	local c = a:Copy()

	if c.suffixes[1] == nil and not c.space then
		local space, short_b = b:match("^(%s+)(.*)$")
		if space then
			c.space = space
			b = short_b
		end
	end
	table.insert(c.suffixes, tostring(b))

	return c
end

function Modifier:__tostring()
	local str = table.concat(self.suffixes)
	if self.space then
		str = str..self.space
	end
	return str..tostring(self.str)
end


local Adjective = Class(Modifier, function(self, str)
	Modifier._ctor(self, str)
end)

local Adverb = Class(Modifier, function(self, str)
	Modifier._ctor(self, str)
end)

function Adverb:__tostring()
	local str = table.concat(self.suffixes)

	local stem, midspace, lastword = str:match("^(..-)(%s+)(%S+)$")
	if stem then
		return stem..(self.space or "")..tostring(self.str)..self.space..lastword
	else
		return Modifier.__tostring(self)
	end
end


--[[
-- Table proxying.
--]]

-- handler takes care of the indexing, receiving the real value indexed.
local function NewTableAccessProxy(handler)
	-- Meta is also the key to the hidden table.
	local meta = {}

	local function bless(t)
		return setmetatable({[meta] = {t = t}}, meta)
	end

	function meta:__index(k)
		local metadata = rawget(self, meta)
		local hidden = assert(metadata.t)
		local v = hidden[k]
		if type(v) == "table" then
			v = bless(v)
			local newbranch
			if metadata.branch then
				newbranch = metadata.branch.."."
			else
				newbranch = ""
			end
			newbranch = newbranch..k
			rawget(v, meta).branch = newbranch
		else
			v = handler(v, metadata.branch, k)
		end
		rawset(self, k, v)
		return v
	end

	function meta:__newindex(k, v)
		rawset(rawget(self, meta), k, v)
	end

	return bless
end

local function modifierWrapper(v, branch, k)
	if type(v) == "string" then
		if branch == "WET_PREFIX" and k == "GOOP" then
			return Adverb(v)
		else
			return Adjective(v)
		end
	else
		return v
	end
end

local StringsProxy = NewTableAccessProxy(modifierWrapper)(_G.STRINGS)

-- Makes a function use StringsProxy instead of STRINGS.
local makeStringsProxiedFunction = (function()
	local env = {
		STRINGS = StringsProxy,
	}

	setmetatable(env, {
		__index = _G,
		__newindex = _G,
	})

	return function(fn)
		_G.setfenv(fn, env)
		return fn
	end
end)()


--[[
-- Binding of the above constructs to the game.
--]]
local oldGetDisplayName = EntityScript.GetDisplayName
if oldGetDisplayName then
	makeStringsProxiedFunction(oldGetDisplayName)

	function EntityScript:GetDisplayName(...)
		local fullname = oldGetDisplayName(self, ...)
		return tostring(fullname)
	end
end


AddComponentPostInit("inspectable", function(Inspectable, inst)
	inst:DoTaskInTime(0, function(inst)
		if inst.wet_prefix then
			if inst.prefab == "wetgoop" then
				inst.wet_prefix = Adverb(inst.wet_prefix)
			else
				inst.wet_prefix = Adjective(inst.wet_prefix)
			end
		end
	end)
end)
